# =============================================================================
# Copyright (c) 2024 by Laszlo Anka. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================
import concurrent.futures
from typing import Any, TYPE_CHECKING, Optional
from abc import abstractmethod
import threading

from pypz.core.channels.base import ChannelBase, ChannelMetric
from pypz.core.channels.status import ChannelStatus
from pypz.core.commons.utils import current_time_millis

if TYPE_CHECKING:
    from pypz.core.specs.plugin import InputPortPlugin, OutputPortPlugin


class ChannelReader(ChannelBase):
    """
    This class is the base of the input channel classes.
    The idea is that the class provides some core, protected abstract methods, which shall be implemented by
    the developer. These implementations then will be invoked by invoker methods. This makes sure that additional
    necessary logic will be performed along the implemented.

    :param channel_name: name of the channel
    :param context: the :class:`PortPlugin <pypz.core.specs.plugin.PortPlugin>`, which operates this channel
    :param executor: an external ThreadPoolExecutor, if not provided, on will be created internally
    """

    # ======================= static fields =======================

    NotInitialized = -1
    """
    Some helper static to init or identify if offset is not initialized
    """

    MetricsBufferLength = 10
    """
    This variable specifies the length of the metric buffer and so the averaging window
    """

    # ======================= Ctor =======================

    def __init__(self, channel_name: str,
                 context: 'InputPortPlugin',
                 executor: Optional[concurrent.futures.ThreadPoolExecutor] = None,
                 **kwargs):
        super().__init__(channel_name, context, executor, **kwargs)

        self._read_record_count: int = 0
        """
        Counter for how many records have been read by this channel. Note that this value can differ from the
        offset value in case of crash-restart. Still the framework maintains the difference between this value
        and the loaded offset and will compensate the framework calculated offset by that difference.
        """

        self._read_record_offset: int = ChannelReader.NotInitialized
        """
        The actual write offset of the read records. Eg. if 10 records have been read and provided to the plugin
        then the offset is 10. Note that this might differ from the m_inputRecordCount_i64 in case of crash-restart.
        """

        self._initial_input_record_offset: int = 0
        """
        This is variable stores the initial offset, which can be different than 0 in case of crash-restart. In this
        case this value is actually the difference by which the framework calculated offset shall be compensated.
        """

        self._last_offset_committed: int = 0
        """
        This value tracks the last committed offset. The main use-case of this variable is to prevent offset commit
        if the new offset equals the last committed.
        """

        self._metrics_buffer: list[ChannelMetric] = list()
        """
        This list holds metric elements, however the only purpose is to be aware of the first
        element to discard like in a circular buffer
        """

        self._aggregated_time_between_reads: int = 0
        """
        Metric value to store the summed up times between reads
        """

        self._aggregated_record_count: int = 0
        """
        Metric value to store the summed up record count
        """

        self._current_read_timestamp: int = 0
        """
        Stores the timestamp of the most current read to be able to calculate elapsed time
        """

        self._current_read_record_count: int = 0
        """
        Stores the record count of the most current read
        """

        self._status_sender_lock = threading.Lock()
        """
        This lock is used to prevent concurrent modification of metrics
        """

    # ======================= Abstract methods =======================

    @abstractmethod
    def _load_input_record_offset(self) -> int:
        """
        This method shall implement the logic to retrieve stored read offset from the underlying technology. This
        can be arbitrary, but the method shall return the stored offset.

        :return: stored offset
        """
        pass

    @abstractmethod
    def has_records(self) -> bool:
        """
        This method shall implement the logic to check, if the reader can still read records.

        :return: True if channel has still records, False if not
        """
        return True

    @abstractmethod
    def _read_records(self) -> list[Any]:
        """
        This method shall implement the logic to read records from the input channel. An ArrayList of records is
        expected.

        :return: list of records OR empty ArrayList if no records read. Null is not accepted.
        """
        pass

    @abstractmethod
    def _commit_offset(self, offset: int) -> None:
        """
        This method shall implement the logic that commits the provided offset using the underlying technology.

        :param offset: provided offset to commit
        """
        pass

    # ======================= Getter/setter/invoker methods =======================

    def get_read_record_count(self) -> int:
        return self._read_record_count

    def get_read_record_offset(self) -> int:
        return self._read_record_offset

    def set_initial_record_offset(self, initial_record_offset: int) -> None:
        """
        This method initializes the internal variables

        :param initial_record_offset: value to be initialized to
        """

        self._initial_input_record_offset = initial_record_offset
        self._last_offset_committed = initial_record_offset
        self._read_record_offset = initial_record_offset

    def set_initial_record_offset_auto(self) -> None:
        """
        This method sets the initial values of the internal variables to the values retrieved by the implementation.
        """

        self.set_initial_record_offset(self._load_input_record_offset())

    def acknowledge_input(self) -> None:
        """
        This method sends the corresponding ack signal to the ChannelWriter
        """

        self.invoke_sync_send_status_message(ChannelStatus.Acknowledged)

    def invoke_read_records(self) -> list[Any]:
        """
        An invoker method that encapsulates the actual implementation. This method
        MUST be called instead of the implemented method directly to ensure proper
        channel functionality.

        :return: list of read records or empty list in case no records read
        """

        read_records = self._read_records()

        read_record_count = len(read_records) if read_records is not None else 0

        if 0 < read_record_count:
            self._read_record_count += read_record_count
            self._read_record_offset += read_record_count
            self._health_check_payload["receivedRecordCount"] = self._read_record_count

        if self._metrics_enabled:
            with self._status_sender_lock:
                if 0 < self._current_read_timestamp:
                    new_metric = ChannelMetric(current_time_millis() - self._current_read_timestamp,
                                               self._current_read_record_count)

                    self._aggregated_time_between_reads += new_metric.elapsedTimeSinceLastIO
                    self._aggregated_record_count += new_metric.recordCountInLastIO

                    self._metrics_buffer.append(new_metric)

                    if ChannelReader.MetricsBufferLength == len(self._metrics_buffer):
                        old_metric = self._metrics_buffer.pop(0)
                        self._aggregated_time_between_reads -= old_metric.elapsedTimeSinceLastIO
                        self._aggregated_record_count -= old_metric.recordCountInLastIO

                self._current_read_timestamp = current_time_millis()
                self._current_read_record_count = read_record_count

        return read_records

    def on_status_message_send(self) -> None:
        if self._metrics_enabled:
            with self._status_sender_lock:
                self._health_check_payload["elapsedTimeSinceLastReadCycleMs"] = 0 if \
                    0 == self._current_read_timestamp else current_time_millis() - self._current_read_timestamp
                self._health_check_payload["averageTimePerReadCycleMs"] = 0 if 0 == len(self._metrics_buffer) else \
                    self._aggregated_time_between_reads / len(self._metrics_buffer)
                self._health_check_payload["averageTimePerReadRecordMs"] = 0 \
                    if 0 == self._aggregated_record_count else \
                    self._aggregated_time_between_reads / self._aggregated_record_count
                self._health_check_payload["averageRecordPerReadCycle"] = 0 if 0 == len(self._metrics_buffer) else \
                    self._aggregated_record_count / len(self._metrics_buffer)

    def invoke_commit_offset(self, offset: int, compensate_with_initial_offset: bool = True) -> None:
        """
        This method is used to invoke the implementation of the abstract method. This is necessary to perform some
        additional actions like compensating the calculated offset with the initial to make sure that the proper
        offset is committed. Note that this compensation makes only sense in case of crash-restart, since the initial
        values are probably not 0. Note that the provided offset is calculated by the calculateApplicableOffsetFrom
        by breaking down the calculated offset by the framework.

        :param offset: offset calculated by the plugin for this channel
        :param compensate_with_initial_offset: if True, the initial offset will be added to the provided offset
        """

        offset_to_commit = offset

        if compensate_with_initial_offset:
            offset_to_commit += self._initial_input_record_offset

            if offset_to_commit > self._read_record_offset:
                raise ValueError(f"Offset is out of range: {offset_to_commit}; "
                                 f"Current read offset? {self._read_record_offset}")

        if offset_to_commit != self._last_offset_committed:
            self._commit_offset(offset_to_commit)

            self._last_offset_committed = offset_to_commit

        if offset_to_commit < self._last_offset_committed:
            self._logger.warning(f"!WARNING! Offset to commit ({offset_to_commit}) "
                                 f"is lower than the last committed offset ({self._last_offset_committed}). "
                                 f"Commit ignored.")

    def invoke_commit_current_read_offset(self) -> None:
        """
        This method is used to invoke the implementation of the abstract method.
        """

        self.invoke_commit_offset(self._read_record_offset, False)


class ChannelWriter(ChannelBase):
    """
    This class is the base of the output channel classes.
    The idea is that the class provides some core, protected abstract methods, which shall be implemented by
    the developer. These implementations then will be invoked by invoker methods. This makes sure that additional
    necessary logic will be performed along the implemented.

    :param channel_name: name of the channel
    :param context: the :class:`PortPlugin <pypz.core.specs.plugin.PortPlugin>`, which operates this channel
    :param executor: an external ThreadPoolExecutor, if not provided, on will be created internally
    """

    MetricsBufferLength = 10
    """
    This variable specifies the length of the metric buffer and so the averaging window
    """

    # ======================= Ctor =======================

    def __init__(self, channel_name: str,
                 context: 'OutputPortPlugin',
                 executor: Optional[concurrent.futures.ThreadPoolExecutor] = None,
                 **kwargs):
        super().__init__(channel_name, context, executor, **kwargs)

        self._written_record_count: int = 0
        """
        Number of outputted records.
        """

        self._metrics_buffer: list[ChannelMetric] = list()
        """
        This list holds metric elements, however the only purpose is to be aware of the first
        element to discard like in a circular buffer
        """

        self._aggregated_time_between_outputs: int = 0
        """
        Metric value to store the summed up times between reads
        """

        self._aggregated_record_count: int = 0
        """
        Metric value to store the summed up record count
        """

        self._current_output_timestamp: int = 0
        """
        Stores the timestamp of the most current read to be able to calculate elapsed time
        """

        self._current_output_record_count: int = 0
        """
        Stores the record count of the most current read
        """

        self._status_sender_lock = threading.Lock()
        """
        This lock is used to prevent concurrent modification of metrics
        """

    # ======================= Abstract methods =======================

    @abstractmethod
    def _write_records(self, records: list[Any]) -> None:
        """
        This method shall implement the logic that writes the records to the output resource. It will automatically
        be invoked by the plugin via the corresponding invoker method.

        :param records: list of records to be written
        """
        pass

    # ======================= Getter/setter/invoker methods =======================

    def get_written_record_count(self):
        return self._written_record_count

    def is_all_connected_input_channels_acknowledged(self) -> bool:
        """
        This method retrieves the number of InputChannels that have acknowledged their input (i.e. the output of
        this channel). This information can be useful for scenarios, where a synchronization logic needs to be
        implemented e.g. where the output shall not provide more data until the inputs did not acknowledge. Note
        that the state is maintained entirely by the ChannelWriter i.e. if the ChannelReader sent the proper message,
        a flag will be set in the corresponding entry of the status map. Then this flag will be reset once this
        ChannelWriter produces new data (check invoke_write_records() for more detail).

        :return: number of acknowledged InputChannels
        """

        for status_monitor in list(self._status_map.values()):
            if status_monitor.is_channel_healthy() and \
                    (ChannelStatus.Acknowledged not in status_monitor.status_update_map):
                return False
        return True

    def invoke_write_records(self, records: list[Any]) -> None:
        """
        This method is used to invoke the implementation of the abstract method. This is necessary to perform some
        additional actions e.g. updating number of outputted records.

        :param records: records to be written
        """

        self._write_records(records)

        output_record_count = len(records)

        self._written_record_count += output_record_count

        self._health_check_payload["sentRecordCount"] = self._written_record_count

        # We need to set the acknowledged flag to false, since we output new data to be acknowledged
        for status_monitor in list(self._status_map.values()):
            if ChannelStatus.Acknowledged in status_monitor.status_update_map:
                del status_monitor.status_update_map[ChannelStatus.Acknowledged]

        if self._metrics_enabled:
            with self._status_sender_lock:
                if 0 < self._current_output_timestamp:
                    new_metric = ChannelMetric(current_time_millis() - self._current_output_timestamp,
                                               self._current_output_record_count)

                    self._aggregated_time_between_outputs += new_metric.elapsedTimeSinceLastIO
                    self._aggregated_record_count += new_metric.recordCountInLastIO

                    self._metrics_buffer.append(new_metric)

                    if ChannelWriter.MetricsBufferLength == len(self._metrics_buffer):
                        old_metric = self._metrics_buffer.pop(0)
                        self._aggregated_time_between_outputs -= old_metric.elapsedTimeSinceLastIO
                        self._aggregated_record_count -= old_metric.recordCountInLastIO

                self._current_output_timestamp = current_time_millis()
                self._current_output_record_count = output_record_count

    def on_status_message_send(self):
        if self._metrics_enabled:
            with self._status_sender_lock:
                self._health_check_payload["elapsedTimeSinceLastOutputCycleMs"] = 0 \
                    if 0 == self._current_output_timestamp else current_time_millis() - self._current_output_timestamp
                self._health_check_payload["averageTimePerOutputCycleMs"] = 0 if 0 == len(self._metrics_buffer) else \
                    self._aggregated_time_between_outputs / len(self._metrics_buffer)
                self._health_check_payload["averageTimePerOutputRecordMs"] = 0 \
                    if 0 == self._aggregated_record_count else \
                    self._aggregated_time_between_outputs / self._aggregated_record_count
                self._health_check_payload["averageRecordPerOutputCycle"] = 0 if 0 == len(self._metrics_buffer) else \
                    self._aggregated_record_count / len(self._metrics_buffer)



# =============================================================================
# Copyright (c) 2024 by Laszlo Anka. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================
from typing import cast

import yaml

from pypz.core.specs.dtos import PipelineInstanceDTO, PipelineSpecDTO
from pypz.core.specs.instance import Instance, RegisteredInterface
from pypz.core.specs.operator import Operator


class Pipeline(Instance[Operator], RegisteredInterface):
    """
    This class represents the pipeline instance specs. A pipeline is actually a
    virtual organization of operators. It has a meaning only on pipeline level
    actions like deployment and execution. A pipeline spec can contain
    operators as nested instance.

    :param name: name of the instance, if not provided, it will be attempted to deduce from the variable's name
    """

    def __init__(self, name: str, *args, **kwargs):
        super().__init__(name, Operator, *args, **kwargs)

    def get_dto(self) -> PipelineInstanceDTO:
        instance_dto = super().get_dto()

        # Replicas must be excluded from the dto
        instance_dto.spec.nestedInstances = {
            operator.get_dto() for operator in self.get_protected().get_nested_instances().values()
            if operator.is_principal()
        }

        return PipelineInstanceDTO(name=instance_dto.name,
                                   parameters=instance_dto.parameters,
                                   dependsOn=instance_dto.dependsOn,
                                   spec=PipelineSpecDTO(**instance_dto.spec.__dict__))

    @staticmethod
    def create_from_string(source, *args, **kwargs) -> 'Pipeline':
        return Pipeline.create_from_dto(PipelineInstanceDTO(**yaml.safe_load(source)), *args, **kwargs)

    @staticmethod
    def create_from_dto(instance_dto: 'PipelineInstanceDTO', *args, **kwargs) -> 'Pipeline':
        return cast(Pipeline, Instance.create_from_dto(instance_dto, *args, **kwargs))

    def _on_interrupt(self, system_signal: int = None) -> None:
        pass

    def _on_error(self) -> None:
        pass



# =============================================================================
# Copyright (c) 2024 by Laszlo Anka. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================
import os
import re
import threading
import time
from typing import Generic, TypeVar, Any


def ensure_type(value, expected_type: type):
    """
    Convenience method to ensure that the value has the expected type.

    :param value: value to check
    :param expected_type: expected type

    :raises: TypeError
    """
    if not isinstance(value, expected_type):
        raise TypeError(f"Expected type {expected_type} got {type(value)} instead")

    return value


def current_time_millis() -> int:
    return int(time.time() * 1000)


ReferenceType = TypeVar('ReferenceType')


class SynchronizedReference(Generic[ReferenceType]):
    """
    This class realizes a synchronized reference i.e., it locks the object during
    read and write to avoid race conditions. This class is intended to safeguard
    pypz for the time after GIL has been removed for good.

    :param reference: the reference to the object with the generic type ReferenceType
    """

    def __init__(self, reference: ReferenceType):
        self.__reference: ReferenceType = reference
        self.__lock = threading.Lock()

    def get(self) -> ReferenceType:
        with self.__lock:
            return self.__reference

    def set(self, reference: ReferenceType) -> None:
        with self.__lock:
            self.__reference = reference


class TemplateResolver:
    """
    This class realizes the logic to resolve templates in strings. As of 02/2024 only
    environment variables can be resolved.

    :param left_template_boundary: the start of the template
    :param right_template_boundary: the end of the template
    """

    def __init__(self, left_template_boundary: str, right_template_boundary: str):

        self.m_templatePattern = re.escape(left_template_boundary) + r'.*?' + re.escape(right_template_boundary)
        """
        Regex to find the template pattern
        """

        self.m_envVarPattern = r'(?<=' + re.escape(left_template_boundary) + r'env:)' \
                               r'(.*?)(?=' + re.escape(right_template_boundary) + r')'
        """
        Regex to find the env var specifier pattern in the template pattern
        """

    def resolve(self, lookup_object):
        """
        This method attempts to recursively resolve template strings in either a Map or Collection
        or a normal String.

        :param lookup_object: input object, where the resolution shall take place
        :return: the modified object
        """

        if isinstance(lookup_object, dict):
            new_map = dict()

            for key in lookup_object.keys():
                new_map[key] = self.resolve(lookup_object[key])

            return new_map
        elif isinstance(lookup_object, list):
            new_list = list()

            for element in lookup_object:
                new_list.append(self.resolve(element))

            return new_list
        elif isinstance(lookup_object, set):
            new_set = set()

            for element in lookup_object:
                new_set.add(self.resolve(element))

            return new_set
        elif isinstance(lookup_object, str):
            lookup_string = str(lookup_object)
            resolved_string = lookup_object

            template_matches = re.findall(self.m_templatePattern, lookup_string)

            for templateMatch in template_matches:
                resolved = None

                env_var_matches = re.findall(self.m_envVarPattern, templateMatch)

                if 0 < len(env_var_matches):
                    resolved = os.getenv(env_var_matches[0])

                resolved_string = resolved_string.replace(templateMatch, "" if resolved is None else resolved)

            return resolved_string
        else:
            return lookup_object


class InterruptableTimer:
    """
    This is a utility timer that can be interrupted in thread safe manner.
    Note that this should not be used in cases, where accuracy matters.
    """

    def __init__(self):
        self.__interrupted: SynchronizedReference[bool] = SynchronizedReference(False)

    def interrupt(self):
        self.__interrupted.set(True)

    def sleep(self, seconds: float):
        millis = round(seconds * 1000.0)
        start_time = current_time_millis()

        while ((not self.__interrupted.get()) and
               (millis >= (current_time_millis() - start_time))):
            pass


def convert_to_dict(obj: Any) -> Any:
    """
    This method attempts to convert an object recursively to a dictionary.

    :param obj: any object that can be converted to dict
    :return: resulted dict or the object, if no valid conversion available (necessary due to the recursion)
    """

    if isinstance(obj, list):
        return [convert_to_dict(item) for item in obj]
    elif isinstance(obj, tuple):
        return tuple(convert_to_dict(item) for item in obj)
    elif isinstance(obj, set):
        return {convert_to_dict(item) for item in obj}
    elif isinstance(obj, dict):
        return {key: convert_to_dict(value) for key, value in obj.items()}
    elif hasattr(obj, '__dict__'):
        return {key: convert_to_dict(value) for key, value in obj.__dict__.items()}
    else:
        return obj


def is_type_allowed(obj, allowed_types: tuple) -> bool:
    """
    This convenience function checks if the type of the provided object is allowed given
    the tuple of allowed types provided as argument.

    :param obj: object to check
    :param allowed_types: allowed types

    :return: True if allowed, False otherwise
    """

    if not isinstance(obj, allowed_types):
        return False

    if isinstance(obj, list):
        return all(is_type_allowed(item, allowed_types) for item in obj)
    elif isinstance(obj, tuple):
        return all(is_type_allowed(item, allowed_types) for item in obj)
    elif isinstance(obj, set):
        return all(is_type_allowed(item, allowed_types) for item in obj)
    elif isinstance(obj, dict):
        return all(is_type_allowed(value, allowed_types) for value in obj.values())
    elif hasattr(obj, '__dict__'):
        return all(is_type_allowed(value, allowed_types) for value in obj.__dict__.values())

    return True



# =============================================================================
# Copyright (c) 2024 by Laszlo Anka. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================
import inspect
import re
import sys
import types
import typing
from abc import ABCMeta, ABC, abstractmethod
from typing import Type, Any, TypeVar, Generic, Optional

import yaml

from pypz.core.commons.utils import TemplateResolver, is_type_allowed, convert_to_dict
from pypz.core.commons.parameters import retrieve_parameters, ExpectedParameter, allowed_param_types
from pypz.core.specs.dtos import InstanceDTO, SpecDTO
from pypz.core.specs.utils import InstanceParameters, AccessWrapper, load_class_by_name, remove_super_classes, \
    IncludedCascadingParameterPrefix, ExcludedCascadingParameterPrefix


class RegisteredInterface:
    """
    This is a marker interface with the sole purpose to identify the native specs interfaces.
    It is necessary to separate the implemented and multiple inherited classes from the
    actual plugin interfaces.
    """
    pass


class InterceptedInstance:
    """
    This class represents an intercepted instance at init. Part of the feature set is that
    if you don't provide a name to an instance in its parent context, then its variable name
    is used as name. However, since there is no convenient way to provide the name of the variable
    holding the reference to the object into the object's __init__, we need to work around this.
    The concept is that we intercept the instance creation, if no name has been provided through
    a metaclass. This metaclass creates an InterceptedInstance object instead of the real
    object. As next step, the Instance's __setattr__ is modified (since that is the only place,
    where the variable name is given) so that, if it sees that an InterceptedInstance is to
    be assigned, it instead creates the original object, but this time with the name and the
    context object already provided. Hence, all the (nested) instances' __init__ will be
    performed with already available name and context.
    """

    def __init__(self, instance_name: str, context_class: Type['Instance'], *args, **kwargs):
        self.instance_name = instance_name
        self.context_class = context_class
        self.args: tuple = args
        self.kwargs: dict = kwargs


InterceptedInstanceType = TypeVar('InterceptedInstanceType')
"""
This type definition is necessary to allow proper typehints
of the objects created by the Instance class with the custom
metaclass. Without this type definition no hints will be given.
"""


class InstanceInitInterceptor(ABCMeta):
    """
    This metaclass has the responsibility to intercept the instance initialization and so
    ensure that instance name and context objects are resolved automatically, before
    executing the __init__ of the instance. For more information refer to InterceptedInstance.
    """

    @typing.no_type_check
    def __call__(cls: Type[InterceptedInstanceType], name: str = None, *args, **kwargs) -> InterceptedInstanceType:
        """
        This method is called, if a class object is being created, hence we can intercept
        the __init__ call. The following rules are considered for interception:
        - caller frame shall be an object and shall be an instance of Instance
        - name or context object is not provided
        In these cases is the Instance init intercepted, in any other cases a normal
        init invoked.

        :param name: instance name
        :param args: args
        :param kwargs: kwargs
        :return: either InterceptedInstance or normal object
        """

        if ("self" in inspect.currentframe().f_back.f_locals) and \
                isinstance(inspect.currentframe().f_back.f_locals["self"], Instance) and \
                ((name is None) or ("context" not in kwargs)):
            return type.__call__(InterceptedInstance, name, cls, *args, **kwargs)

        if (name is None) or not re.match(r"^[a-zA-Z_]+[a-zA-Z0-9_-]*$", name):
            raise AttributeError(f"Invalid instance name: {name}; "
                                 f"Name must match pattern: ^[a-zA-Z0-9_]+[a-zA-Z0-9_-]*$")

        instance = type.__call__(cls, name, *args, **kwargs)
        instance.__on_init_finished__(*args, **kwargs)
        return instance


NestedInstanceType = TypeVar('NestedInstanceType', bound='Instance')


class Instance(Generic[NestedInstanceType], RegisteredInterface, ABC, metaclass=InstanceInitInterceptor):
    """
    This class represents the instance specs that serves as base for
    all the other models. It abstracts all common features and functionalities
    that is required to maintain the actual instances like parameters
    or even multi level nested instances.

    :param name: name of the instance, if not provided, it will be attempted to deduce from the variable's name
    :param nested_instance_type: type of the expected nested instances
    """

    # ========= ctor ==========

    def __init__(self, name: str = None,
                 nested_instance_type: Type[NestedInstanceType] = None,
                 *args, **kwargs):

        # Sanity checks
        # =============

        if name is not None and not isinstance(name, str):
            raise TypeError(f"Invalid instance name: {name} {type(name)}")

        if nested_instance_type is not None and not issubclass(nested_instance_type, Instance):
            raise TypeError(f"Invalid nested instance type: {nested_instance_type}. "
                            f"Must be an extension of {Instance}")

        # Member declarations
        # ===================

        self.__nested_instance_type: Optional[Type[NestedInstanceType]] = nested_instance_type
        """
        Stores the specified type of the nested instances. This is required to
        be able to discover those instances
        """

        self.__reference = kwargs["reference"] if "reference" in kwargs else None
        """
        Reference to the reference instance. If specified, then some instance internal
        configuration related attributes will refer to the attributes of the reference instance.
        """

        self.__context: Instance = kwargs["context"] if "context" in kwargs else None
        """
        Reference to the context aka parent object. Derived automatically
        in the override implementation of __setattr__.
        """

        self.__nested_instances: dict[str, NestedInstanceType] = dict()
        """
        This dictionary holds all the instances that are nested in the context
        of this instance object
        """

        self.__simple_name: str = name
        """
        Name of the instance, which is represented by the object created from the
        implementation class. If not provided and there is a parent context, then
        the parent context will use the name of the variable.
        """

        self.__full_name: str = self.__simple_name \
            if self.__context is None else self.__context.get_full_name() + "." + self.__simple_name
        """
        Full name of the instance which uniquely identifies it up to its topmost context.
        For example, if an instance A has a parent context B, which has a parent context C,
        then the full name of A is 'C.B.A'. This value is calculated and the result is
        stored after the first calculation to avoid recalculation every time.
        """

        self.__spec_name: str = ":".join([self.__class__.__module__, self.__class__.__qualname__])
        """
        The name of the spec constructed of module and qualified class name. Notice that we separate
        the module name and the class name so that we can identify them at loading by name.
        """

        self.__spec_classes: set = set([
            spec_class for spec_class in self.__class__.__mro__
            if issubclass(spec_class, Instance) and (RegisteredInterface in spec_class.__bases__)
        ])
        """
        Set of specs classes that are in the class hierarchy i.e., which specs classes are
        contributing to the implementation of this class
        """

        self.__expected_parameters: dict[str, ExpectedParameter] = retrieve_parameters(self, ExpectedParameter)
        """
        Map of expected parameters defined as descriptor of the class. Key is the name of the parameter,
        value is the parameter descriptor. Used to check, if an expected (described) parameter value
        shall be set upon instance parameter setting.
        """

        self.__parameters: InstanceParameters = self.__reference.__parameters \
            if self.__reference is not None else InstanceParameters()
        """
        The interpreted instance parameters i.e., cascading and templates are interpreted
        """

        self.__depends_on: set = self.__reference.__depends_on if self.__reference is not None else set()
        """
        Set of other instances that is this instance depending on. Note however that
        the type of the dependencies are checked dynamically in runtime, since dependencies
        can only be defined on the same type of instance.
        """

        """ Registering expected parameter update. This means that, if a parameter has been set via
        the set_parameter() method, we need to make sure that the corresponding parameter object
        is updated as well. """
        for param in self.__expected_parameters.values():
            self.__parameters.on_parameter_update(param.name,
                                                  lambda value, p=param, instance=self: p.__set__(instance, value))

        """ It is clear that in python anything can be accessed, however methods and attributes that
            should not be used are hidden (at least from autocompletion) from the user, still with
            the following dynamically created method it is possible to access them so that the IDE
            is not complaining. """
        object.__setattr__(self, "get_protected", lambda this=self: AccessWrapper(this))

    # ========= overridable methods ==========

    @abstractmethod
    def _on_interrupt(self, system_signal: int = None) -> None:
        """
        This method can be implemented to react to interrupt signals like
        SIGINT, SIGTERM etc. The specs implementation can then execute interrupt
        logic e.g., early termination of loops.

        :param system_signal: id of the system signal that causes interrupt
        """
        pass

    @abstractmethod
    def _on_error(self) -> None:
        """
        This method can be implemented to react to error events during
        execution. The error itself may come from arbitrary sources.
        """
        pass

    # ========= public methods ==========

    def get_context(self):
        return self.__context

    def get_simple_name(self) -> str:
        return self.__simple_name

    def get_full_name(self) -> str:
        return self.__full_name

    def get_parameter(self, name: str):
        return self.__parameters[name]

    def has_parameter(self, name: str) -> bool:
        return name in self.__parameters

    def set_parameter(self, name: str, value: Any) -> None:
        """
        Parameter setter method, which interprets the templates and handles cascading
        parameters. One can ignore setting values for existing parameters via the
        provided flag.

        :param name: parameter name
        :param value: parameter value
        """

        resolved_value = TemplateResolver("${", "}").resolve(value)

        if not is_type_allowed(resolved_value, allowed_param_types):
            raise TypeError(f"Invalid parameter value type for '{name}': {type(resolved_value)}. "
                            f"Allowed types: {allowed_param_types}")

        if name.startswith(ExcludedCascadingParameterPrefix):
            """ Cascade the parameter further. Notice that the first leading
                cascaded specifier has been removed in the parameter name
                Notice that we copy the instances into a list, since there can be
                situations, where the list is modified during iteration via
                replication (e.g., replicationFactor cascaded from pipeline) """
            for nested_instance in list(self.__nested_instances.values()):
                nested_instance.set_parameter(name[1:], resolved_value)
        else:
            param_name = name
            if name.startswith(IncludedCascadingParameterPrefix):
                """ Remove all the cascading specifier from the parameter name """
                param_name = name \
                    .replace(IncludedCascadingParameterPrefix, "") \
                    .replace(ExcludedCascadingParameterPrefix, "")

                """ Cascade the parameter further. Notice that the first leading
                    cascaded specifier has been removed in the parameter name """
                for nested_instance in self.__nested_instances.values():
                    nested_instance.set_parameter(name[1:], resolved_value)

            self.__parameters[param_name] = resolved_value

    def set_parameters(self, parameters: dict) -> None:
        """
        Convenience parameter setter method for dicts, where all key value pair will be
        set separately

        :param parameters: parameter dict
        """

        for name, value in parameters.items():
            self.set_parameter(name, value)

    def get_expected_parameters(self) -> dict | str:
        """
        Returns all the expected parameters as dictionary. Each parameter has the following types:

        .. code-block:: python

           result = {
               'name': {
                   'type': 'str | int | float | set | list | dict | type(None)',
                   'required': 'True | False',
                   'description': 'str',
                   'currentValue': 'str | int | float | set | list | dict | None'
               }
           }
        """

        expected_parameters = dict()
        [expected_parameters.update(param.to_dict(self))
         for param in self.__expected_parameters.values()]

        return expected_parameters

    def get_missing_required_parameters(self) -> dict[str, set[str]]:
        """
        This method returns the missing required parameters recursively
        for all nested instances, hence all parameters will be collected
        in the current scope.

        :return: dict, where key is the instance's full name and value is
                 a list of the names of the missing required parameters. If no missing
                 required parameters, then an empty dict will be returned
        """

        missing_required_parameters = dict()

        # Calling for nested instances recursively
        for instance in self.__nested_instances.values():
            missing_required_parameters.update(instance.get_missing_required_parameters())

        for expected_param in self.__expected_parameters.values():
            if expected_param.required:
                if (expected_param.name not in self.__parameters) or (self.__parameters[expected_param.name] is None):
                    if self.__full_name not in missing_required_parameters:
                        missing_required_parameters[self.__full_name] = set()
                    missing_required_parameters[self.__full_name].add(expected_param.name)

        return missing_required_parameters

    def depends_on(self, instance: 'Instance') -> None:
        """
        Specify dependency instances of the actual instance. The following
        prerequisites shall be considered:
        - dependency must be the same specs type
        - dependency can be defined only in the same parent context
        The same parent context means that if there is an operatorA with plugin1
        and plugin2 and an operatorB with plugin3, then only plugin1 and plugin2
        can express dependencies to each other, since plugin3 is in another operator.

        :param instance: dependency instance
        """

        if not isinstance(instance, Instance):
            raise TypeError(f"[{self.__full_name}] Invalid dependency type: {type(instance)}")

        # Instance cannot depend on self
        if self is instance:
            raise AttributeError(f"[{self.__full_name}] Invalid dependency. Instance cannot depend on itself. ")

        # Makes no sense to express dependency between instances in different context
        if (self.__context is None) or (instance.__context is None) or \
                (self.__context is not instance.__context):
            raise AttributeError(f"[{self.__full_name}] Invalid dependency. "
                                 f"Dependencies must have identical parent context.")

        # Check circular dependency
        if self in instance.__depends_on:
            raise RecursionError(f"Circular dependency between detected: "
                                 f"{self.__full_name} <-> {instance.__full_name} ")

        self.__depends_on.add(instance)

    def get_dto(self) -> InstanceDTO:
        """
        Converts the instance information into the corresponding Data Transfer Object (DTO)

        :return: DTO from instance
        """

        """ This is necessary to avoid the serialization of the extended dict's state and
            to prevent issues, if the parameters on the DTO are altered """
        raw_parameters = dict()
        raw_parameters.update(self.__parameters)

        return InstanceDTO(name=self.__simple_name,
                           parameters=raw_parameters,
                           dependsOn=[instance.get_simple_name() for instance in self.__depends_on],
                           spec=SpecDTO(name=self.__spec_name,
                                        types=[str(spec_class) for spec_class in
                                               remove_super_classes(self.__spec_classes)],
                                        nestedInstanceType=None if self.__nested_instance_type is None
                                        else str(self.__nested_instance_type),
                                        expectedParameters=self.get_expected_parameters(),
                                        nestedInstances=[instance.get_dto() for instance in
                                                         self.__nested_instances.values()]))

    def update(self, source: InstanceDTO | dict | str) -> None:
        """
        This method allows to update certain attributes of the instance based on the
        provided DTO. All attributes can be updated here as well, for which a setter
        exists, in addition one can inject additional nested instances or update existing.

        :param source: json string, dict or DTO
        """

        if isinstance(source, str):
            instance_dto = InstanceDTO(**yaml.safe_load(source))
        elif isinstance(source, dict):
            instance_dto = InstanceDTO(**source)
        elif isinstance(source, InstanceDTO):
            instance_dto = source
        else:
            raise TypeError(f"Invalid update source type: {type(source)}")

        if (instance_dto.name is not None) and (instance_dto.name != self.__simple_name):
            raise ValueError(f"Mismatching instance name provided ({instance_dto.name}) and actual "
                             f"({self.__simple_name}) instance name. ")

        if instance_dto.parameters is not None:
            """ Note that at this point all the nested instances had to be created, hence we update
                the parameters of this instance before the nested instances' so the concept of
                proximity based precedence of parameter setting is honored. """
            self.set_parameters(instance_dto.parameters)

        if instance_dto.spec is not None:
            if (instance_dto.spec.name is not None) and (instance_dto.spec.name != self.__spec_name):
                raise AttributeError(f"[{self.__full_name}] Mismatching specs name. Loaded: {self.__spec_name}; "
                                     f"Provided: {instance_dto.spec.name}")

            if (instance_dto.spec.nestedInstances is not None) and (0 < len(instance_dto.spec.nestedInstances)):
                if self.__nested_instance_type is None:
                    raise AttributeError(f"[{self.get_full_name()}] No nested instance is expected for {type(self)}")

                for nested_instance_dto in instance_dto.spec.nestedInstances:
                    if nested_instance_dto.name is None:
                        raise ValueError("Missing instance name")

                    if nested_instance_dto.name not in self.__nested_instances:
                        raise AttributeError(f"[{self.get_full_name()}] Instance not found with name: "
                                             f"{nested_instance_dto.name}")

                    self.__nested_instances[nested_instance_dto.name].update(nested_instance_dto)

        if instance_dto.dependsOn is not None:
            for instance_name in instance_dto.dependsOn:
                if instance_name not in self.__context.__nested_instances:
                    raise AttributeError(f"[{self.__full_name}] Instance not found in context "
                                         f"'{self.__context.__full_name}': {instance_name}")

                self.depends_on(self.__context.__nested_instances[instance_name])

    # ========= protected methods ==========

    def __on_init_finished__(self, *args, **kwargs) -> None:
        """
        This method can be overridden to implement logic that shall be executed after
        the Instance's __init__ has been finished.
        """

        instance_dto = kwargs["from_dto"] if "from_dto" in kwargs else None

        if instance_dto is not None:
            # Creating not existing nested instances
            if instance_dto.spec is not None:
                if (instance_dto.spec.name is not None) and (instance_dto.spec.name != self.__spec_name):
                    raise AttributeError(f"[{self.__full_name}] Mismatching specs name: {self.__spec_name}; "
                                         f"Expected: {instance_dto.spec.name}")

                if (instance_dto.spec.nestedInstances is not None) and (0 < len(instance_dto.spec.nestedInstances)):
                    if self.__nested_instance_type is None:
                        raise AttributeError(f"[{self.__full_name}] "
                                             f"No nested instance is expected for {type(self)}")

                    for nested_instance_dto in instance_dto.spec.nestedInstances:
                        if nested_instance_dto.name is None:
                            raise ValueError("Missing instance name")

                        # First we need to make sure that all instances are created.
                        # This has to happen before updating the instances, since if
                        # an already existing instance refers to a not yet existing
                        # then error will arise.
                        if nested_instance_dto.name not in self.__nested_instances:
                            nested_instance_reference = self.__reference.__nested_instances[nested_instance_dto.name] \
                                if self.__reference is not None else None
                            new_nested_instance = self.__nested_instance_type.create_from_dto(
                                nested_instance_dto, context=self, reference=nested_instance_reference,
                                mock_nonexistent=True, disable_auto_update=True)

                            if not isinstance(new_nested_instance, self.__nested_instance_type):
                                raise AttributeError(f"[{new_nested_instance.get_full_name()}] "
                                                     f"Mismatching nested instance type. Expected: "
                                                     f"{self.__nested_instance_type}")
                            self.__setattr__(new_nested_instance.__simple_name, new_nested_instance)

    # ========= internal methods ==========

    def __setattr__(self, name, value):
        """
        Overridden method to allow the discovery, registration and linking of
        nested objects. It checks, whether the value is an instance of the
        expected nested instance type, then it retrieves the instance's
        simple name if not given, and then it links the both the parent and
        the nested context to each other

        :param name: name of the variable
        :param value: value of the variable
        """

        if isinstance(value, InterceptedInstance):
            final_instance_name = value.instance_name if value.instance_name is not None else name
            nested_instance_reference = self.__reference.__nested_instances[final_instance_name] \
                if self.__reference is not None else None
            if (self.__nested_instance_type is not None) and \
                    (issubclass(value.context_class, self.__nested_instance_type)):
                instance = value.context_class(final_instance_name, context=self, reference=nested_instance_reference,
                                               *value.args,  **value.kwargs)
            else:
                instance = value.context_class(final_instance_name, context=None, reference=nested_instance_reference,
                                               *value.args, **value.kwargs)
        else:
            instance = value

        object.__setattr__(self, name, instance)

        if hasattr(self, f"_{Instance.__name__}__nested_instance_type") and \
                (self.__nested_instance_type is not None) and \
                isinstance(instance, self.__nested_instance_type) and \
                (instance is not self.__context):
            if instance.get_simple_name() in self.__nested_instances:
                raise AttributeError(f"Instance with name already declared: {instance.get_simple_name()}")

            self.__nested_instances[instance.get_simple_name()] = instance

    def __eq__(self, other):
        if self is other:
            return True

        return isinstance(other, type(self)) and \
            (self.__simple_name == other.__simple_name) and \
            (self.__parameters == other.__parameters) and \
            (self.__depends_on == other.__depends_on) and \
            (self.__nested_instances == other.__nested_instances) and \
            (self.__expected_parameters == other.__expected_parameters) and \
            (self.__spec_name == other.__spec_name)

    def __str__(self):
        return yaml.safe_dump(convert_to_dict(self.get_dto()), default_flow_style=False)

    def __hash__(self):
        return hash((self.__full_name, self.__spec_name))

    def __getattr__(self, name):
        return self.__getattribute__(name)

    # ================= static methods =====================

    @staticmethod
    def create_from_dto(instance_dto: InstanceDTO, *args, **kwargs) -> 'Instance':
        """
        Creates an instance object from the DTO representation. It is capable
        to retrieve and load specified classes and to update created instances
        according to the DTO.

        :param instance_dto: instance DTO
        :return: instance object specified by the DTO
        """

        if not isinstance(instance_dto, InstanceDTO):
            raise TypeError(f"Invalid instance DTO type: {type(instance_dto)}")

        if instance_dto.name is None:
            raise ValueError("Missing instance name")

        if (instance_dto.spec is None) or (instance_dto.spec.name is None):
            raise AttributeError("Invalid or missing 'specs' definition")

        if 0 > instance_dto.spec.name.find(":"):
            raise ValueError(f"Class loading error. Invalid class specifier: {instance_dto.spec.name}; "
                             f"Class specifier must be composed of [PACKAGE_NAME.]MODULE_NAME:CLASS_NAME[.CLASS_NAME+]")

        if instance_dto.spec.location is not None:
            # TODO - remote retrieval shall be implemented here
            raise NotImplementedError("Location based retrieval not yet implemented")

        mock_nonexistent = ("mock_nonexistent" in kwargs) and (kwargs["mock_nonexistent"])

        try:
            instance: Instance = \
                load_class_by_name(instance_dto.spec.name.replace(":", "."))(instance_dto.name,
                                                                             from_dto=instance_dto,
                                                                             *args, **kwargs)
        except (ModuleNotFoundError, AttributeError) as e:
            if mock_nonexistent:
                if (instance_dto.spec.types is None) or (0 == len(instance_dto.spec.types)):
                    raise AttributeError("Missing type specification: instance.spec.types: List[Str]")

                base_classes = set()
                abstract_methods = set()

                # Retrieve and check base classes specified in spec.types
                for class_name in instance_dto.spec.types:
                    loaded_class = load_class_by_name(class_name)
                    if not issubclass(loaded_class, (Instance, RegisteredInterface)):
                        raise TypeError(f"Invalid spec type: {loaded_class}. Expected subclass of {Instance}")
                    base_classes.add(loaded_class)
                    abstract_methods.update(loaded_class.__abstractmethods__)

                nested_instance_type = None if instance_dto.spec.nestedInstanceType is None \
                    else load_class_by_name(instance_dto.spec.nestedInstanceType)

                # Creating dummy implementation of abstract methods
                class_body: dict[str, Any] = {abstract_method: lambda: None for abstract_method in abstract_methods}

                module_and_name = instance_dto.spec.name.split(":")
                class_body["__module__"] = module_and_name[0]
                class_body["mocked"] = True

                # Create new type with name and bases
                instance_type = types.new_class(module_and_name[1],
                                                tuple(remove_super_classes(base_classes)),
                                                {}, lambda ns: ns.update(class_body))

                if "nested_instance_type" in inspect.signature(instance_type.__init__).parameters:  # type: ignore
                    instance = instance_type(instance_dto.name,
                                             nested_instance_type=nested_instance_type,
                                             from_dto=instance_dto, *args, **kwargs)
                else:
                    instance = instance_type(instance_dto.name, from_dto=instance_dto, *args, **kwargs)

                print(f"[WARNING] Mock instance created of spec '{instance_dto.spec.name}' for "
                      f"[{instance.get_full_name()}]. Reason: {e}", file=sys.stderr)
            else:
                raise

        if ("disable_auto_update" not in kwargs) or (not kwargs["disable_auto_update"]):
            instance.update(instance_dto)

        return instance

    @staticmethod
    def create_from_string(source, *args, **kwargs) -> 'Instance':
        """
        Helper method to provide the functionality to create an instance from
        a json model specified either as string or as dict.

        :param source: model as string
        :return: instance object specified by the DTO
        """
        return Instance.create_from_dto(InstanceDTO(**yaml.safe_load(source)), *args, **kwargs)


class InstanceGroup(ABC):
    """
    This class provides methods to access instance group related information. It can be
    implemented on different instance level, since a group might have different meaning
    on top level and on nested level.
    """

    @abstractmethod
    def get_group_size(self) -> int:
        pass

    @abstractmethod
    def get_group_index(self) -> int:
        pass

    @abstractmethod
    def get_group_name(self) -> Optional[str]:
        pass

    @abstractmethod
    def get_group_principal(self) -> Optional['Instance']:
        pass

    @abstractmethod
    def is_principal(self) -> bool:
        pass



# =============================================================================
# Copyright (c) 2024 by Laszlo Anka. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================
import enum
import time
from abc import ABC, abstractmethod
from typing import Callable, Optional

from pypz.core.specs.operator import Operator
from pypz.core.specs.pipeline import Pipeline
from pypz.executors.commons import ExecutionMode


class DeploymentState(enum.Enum):
    Open = "Open",
    Running = "Running",
    Completed = "Completed",
    Failed = "Failed",
    Unknown = "Unknown",
    NotExisting = "NotExisting"


class Deployer(ABC):
    """
    This is the base class for every deployer implementations. If you want to
    develop your own deployer with your own choice of technology, then you need
    to implement this interface.
    """

    # ================== abstract methods ====================

    @abstractmethod
    def deploy(self, pipeline: Pipeline,
               execution_mode: ExecutionMode = ExecutionMode.Standard,
               ignore_operators: list[Operator] = None,
               wait: bool = True) -> None:
        """
        Shall implement the logic to deploy a pipeline by its instance.

        :param pipeline: pipeline instance to be deployed
        :param execution_mode: execution mode of the operators (check ExecutionMode for details)
        :param ignore_operators: list of operator instance to be excluded from the deployment
        :param wait: True - block until completion; False - don't block until completion
        """

        pass

    @abstractmethod
    def destroy(self, pipeline_name: str, force: bool = False, wait: bool = True) -> None:
        """
        Shall implement the logic to destroy a pipeline by its name.

        :param pipeline_name: name of the deployed pipeline entity
        :param force: True - kill without grace period; False - with grace period
        :param wait: True - block until completion; False - don't block until completion
        """

        pass

    @abstractmethod
    def restart_operator(self, operator_full_name: str, force: bool = False, wait: bool = True) -> None:
        """
        Shall implement the logic to restart a single operator in a pipeline by its name.
        If the operator does not exist, it shall rather create it without throwing an exception.

        :param operator_full_name: full name of the deployed operator
        :param force: True - kill without grace period; False - with grace period
        :param wait: True - block until completion; False - don't block until completion
        """

        pass

    @abstractmethod
    def destroy_operator(self, operator_full_name: str, force: bool = False, wait: bool = True) -> None:
        """
        Shall implement the logic to destroy a single operator by its name.

        :param operator_full_name: full name of the deployed operator
        :param force: True - kill without grace period; False - with grace period
        :param wait: True - block until completion; False - don't block until completion
        """

        pass

    @abstractmethod
    def is_deployed(self, pipeline_name: str) -> bool:
        """
        Shall implement the logic to check, if a pipeline with the specified name has been deployed.

        :param pipeline_name: name of the pipeline to check
        :return: True, if deployed, False if not
        """

        pass

    @abstractmethod
    def retrieve_pipeline_deployments(self) -> set[str]:
        """
        Shall implement the logic to get the names of all deployed pipelines.

        :return: set of names of the deployed pipelines
        """
        pass

    @abstractmethod
    def retrieve_deployed_pipeline(self, pipeline_name: str) -> Optional[Pipeline]:
        """
        Shall implement the logic to retrieve and create the deployed pipeline instance by its name.

        :param pipeline_name: name of the pipeline
        :return: Pipeline object, if existing, None if not existing
        """
        pass

    @abstractmethod
    def retrieve_operator_state(self, operator_full_name: str) -> DeploymentState:
        """
        Shall implement the logic to retrieve the state of a single operator by its name.

        :param operator_full_name: full name of the deployed operator
        :return: check DeploymentState for details
        """
        pass

    @abstractmethod
    def retrieve_operator_logs(self, operator_full_name: str) -> Optional[str]:
        """
        Shall implement the logic to retrieve the logs from a deployed operator by its name.

        :param operator_full_name: full name of the deployed operator
        :return: operator logs as string
        """
        pass

    # ================== public methods ====================

    def retrieve_pipeline_state(self, pipeline_name: str) -> dict[str, DeploymentState]:
        """
        This method retrieves and collects all the operators' states in the deployed pipeline.

        :param pipeline_name: name of the deployed pipeline entity
        :return: a dict, where key is the name of the operator and the value is the corresponding state object
        """

        pipeline: Pipeline = self.retrieve_deployed_pipeline(pipeline_name)

        operator_states = dict()
        for operator in pipeline.get_protected().get_nested_instances().values():
            operator_states[operator.get_full_name()] = self.retrieve_operator_state(operator.get_full_name())

        return operator_states

    def is_any_operator_in_state(self, pipeline_name: str, *state: DeploymentState):
        """
        This method checks, if **any** of the operators in the deployed pipeline is in **any** of the
        specified states.

        :param pipeline_name: name of the deployed pipeline entity
        :param state: list of states in OR condition
        """

        pipeline: Pipeline = self.retrieve_deployed_pipeline(pipeline_name)
        for operator in pipeline.get_protected().get_nested_instances().values():
            operator_state = self.retrieve_operator_state(operator.get_full_name())
            if operator_state in state:
                return True
        return False

    def is_all_operator_in_state(self, pipeline_name: str, *state: DeploymentState):
        """
        This method checks, if **all** the operators in the deployed pipeline is in **any** of the
        specified states.

        :param pipeline_name: name of the deployed pipeline entity
        :param state: list of states in OR condition
        """

        pipeline: Pipeline = self.retrieve_deployed_pipeline(pipeline_name)
        for operator in pipeline.get_protected().get_nested_instances().values():
            operator_state = self.retrieve_operator_state(operator.get_full_name())
            if operator_state not in state:
                return False
        return True

    def attach(self, pipeline_name: str,
               on_operator_state_change: Callable[[Operator, DeploymentState], None] = None) -> None:
        """
        This method attaches itself to a deployed pipeline and remains attached until the pipeline is
        not finished. It is possible to specify callback functions to hook into certain state
        changes. If that state change happens, then the callback gets the Operator instance and
        the corresponding state provided.

        :param pipeline_name: name of the deployed pipeline entity
        :param on_operator_state_change: callback to hook into state changes
        """
        pipeline: Pipeline = self.retrieve_deployed_pipeline(pipeline_name)

        operator_states: dict[Operator, DeploymentState] = \
            {operator: DeploymentState.Unknown for operator in pipeline.get_protected().get_nested_instances().values()}

        finished: bool = False

        while not finished:
            for operator in pipeline.get_protected().get_nested_instances().values():
                operator_state = self.retrieve_operator_state(operator.get_full_name())

                if operator_state != operator_states[operator]:
                    operator_states[operator] = operator_state
                    if on_operator_state_change is not None:
                        on_operator_state_change(operator, operator_state)

            if any((DeploymentState.Open == state) or
                   (DeploymentState.Running == state) for state in operator_states.values()):
                time.sleep(2)
            else:
                finished = True



# =============================================================================
# Copyright (c) 2024 by Laszlo Anka. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================
from typing import Optional, Any, TypeVar

from pypz.core.specs.instance import Instance
from pypz.core.specs.operator import Operator
from pypz.core.specs.plugin import Plugin, PortPlugin, InputPortPlugin, OutputPortPlugin, ResourceHandlerPlugin, \
    ServicePlugin


BlankNestedInstanceType = TypeVar('BlankNestedInstanceType', bound='BlankInstance')


class BlankInstance(Instance[BlankNestedInstanceType]):

    def _on_interrupt(self, system_signal: int = None) -> None:
        pass

    def _on_error(self) -> None:
        pass


class BlankPlugin(BlankInstance[None]):
    pass


class BlankResourceHandlerPlugin(ResourceHandlerPlugin, BlankPlugin):

    def _on_resource_creation(self) -> bool:
        return True

    def _on_resource_deletion(self) -> bool:
        return True


class BlankPortPlugin(PortPlugin, BlankPlugin):

    def _on_port_open(self) -> bool:
        return True

    def _on_port_close(self) -> bool:
        return True


class BlankInputPortPlugin(InputPortPlugin, BlankPortPlugin):

    def can_retrieve(self) -> bool:
        return False

    def retrieve(self) -> Any:
        pass

    def commit_current_read_offset(self) -> None:
        pass


class BlankOutputPortPlugin(OutputPortPlugin, BlankPortPlugin):

    def send(self, data: Any) -> Any:
        pass


class BlankServicePlugin(ServicePlugin, BlankPlugin):

    def _on_service_start(self) -> bool:
        return True

    def _on_service_shutdown(self) -> bool:
        return True


class BlankOperator(Operator, BlankInstance[Plugin]):

    def __init__(self, name: str = None, replication_factor: int = None, *args, **kwargs):
        super().__init__(name, replication_factor, *args, **kwargs)

    def _on_init(self) -> bool:
        return True

    def _on_running(self) -> Optional[bool]:
        return True

    def _on_shutdown(self) -> bool:
        return True



# =============================================================================
# Copyright (c) 2024 by Laszlo Anka. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================
import concurrent.futures
import signal
from typing import Optional

from pypz.executors.commons import ExecutionMode
from pypz.executors.operator.executor import OperatorExecutor
from pypz.core.specs.pipeline import Pipeline


class PipelineExecutor:
    """
    This class implements the feature to execute an entire pipeline i.e., each operator
    individually via an :class:`OperatorExecutor <pypz.executors.operator.executor.OperatorExecutor>`.
    To increase the performance of the execution, each
    :class:`OperatorExecutor <pypz.executors.operator.executor.OperatorExecutor>` will be launched
    in a separate thread.

    .. warning::
       This class has the sole purpose to allow you to test pipelines locally, it should not be used
       for production workloads, since python is not very optimal for concurrent executions. That is
       the reason to limit the operator count.

    :param pipeline: the pipeline instance to be executed
    """

    _max_operator_count: int = 32
    """
    This value limits the number of operators to be executed parallely. An exception will be
    thrown, if exceeded.
    """

    def __init__(self, pipeline: Pipeline):
        signal.signal(signal.SIGTERM, self.interrupt)
        signal.signal(signal.SIGINT, self.interrupt)

        self.__pipeline: Pipeline = pipeline

        self.__executor: Optional[concurrent.futures.ThreadPoolExecutor] = None

        self.__operator_executors: set[OperatorExecutor] = set()

        self.__futures: set[concurrent.futures.Future] = set()

        """ Creating the OperatorExecutor objects. Notice that none of the OperatorExecutors
            may handle interrupts, since this will be handled on PipelineExecutor level. """
        for operator in self.__pipeline.get_protected().get_nested_instances().values():
            self.__operator_executors.add(OperatorExecutor(operator, handle_interrupts=False))

        if PipelineExecutor._max_operator_count < len(self.__operator_executors):
            raise AttributeError(f"Max number of operators exceeded "
                                 f"({PipelineExecutor._max_operator_count}): "
                                 f"{len(self.__operator_executors)}")

    def start(self, exec_mode: ExecutionMode = ExecutionMode.Standard):
        """
        This method triggers the execution by creating a ``ThreadPoolExecutor`` and submitting
        the :class:`OperatorExecutor <pypz.executors.operator.executor.OperatorExecutor>`'s
        corresponding method.

        :param exec_mode: :class:`pypz.executors.commons.ExecutionMode`
        """

        if self.__executor is None:
            self.__executor = concurrent.futures.ThreadPoolExecutor(max_workers=PipelineExecutor._max_operator_count,
                                                                    thread_name_prefix=self.__class__.__name__)
            self.__futures.clear()
            for operator_executor in self.__operator_executors:
                self.__futures.add(self.__executor.submit(operator_executor.execute, exec_mode))

            while any(not future.done() for future in self.__futures):
                pass

    def shutdown(self):
        """
        This method shuts down the ``ThreadPoolExecutor``. Notice that it blocks until all
        OperatorExecutor has finished. Notice as well that we don't cancel futures, since
        it will be handled upon handling the interrupt signals.
        """

        if self.__executor is not None:
            self.__executor.shutdown(wait=True, cancel_futures=False)
            self.__executor = None

    def interrupt(self, signal_number, current_stack):
        """
        This method is called upon receiving a system signal e.g., SIGINT.
        We are interrupting each :class:`OperatorExecutor <pypz.executors.operator.executor.OperatorExecutor>`
        by invoking :meth:`interrupt() <pypz.executors.operator.executor.OperatorExecutor.interrupt>` only
        if it is still running. Notice that we cancel futures i.e., should an
        :class:`OperatorExecutor <pypz.executors.operator.executor.OperatorExecutor>` not yet be scheduled,
        it will prevent to be scheduled.
        """

        for operator_executor in self.__operator_executors:
            if operator_executor.is_running():
                operator_executor.interrupt(signal_number, current_stack)
        for future in self.__futures:
            future.cancel()



# =============================================================================
# Copyright (c) 2024 by Laszlo Anka. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================
from enum import Enum


class ExecutionMode(Enum):

    Skip = "Skip"
    """
    If this mode is set, then the executor will not be invoked. This is necessary
    e.g., in deployment cases, where we want to ignore the execution of certain operators.
    """

    Standard = "Standard"
    """
    In this mode the state machine will follow its standard route.
    Note that `Standard` means that resources will be created and deleted up automatically. If
    you want to have a standard mode, where resources are only created, but not deleted, then use the
    WithoutResourceDeletion
    """

    WithoutResourceDeletion = "WithoutResourceDeletion"
    """
    In this mode the state machine will follow its standard route.
    However the StateResourceDeletion will be skipped. This is a relevant use-case,
    if you want to trigger centralized resource cleanup after the operations are done.
    """

    ResourceCreationOnly = "ResourceCreationOnly"
    """
    In this mode the state machine will execute resource creation only. Note that services
    will be started before and stopped after resource creation.
    """

    ResourceDeletionOnly = "ResourceDeletionOnly"
    """
    In this mode the state machine will execute resource deletion only. Note that services
    will be started before and stopped after resource creation.
    """


class ExitCodes(Enum):

    # These are the common error codes
    # ================================

    NoError = 0
    GeneralError = 1
    CommandCannotBeExecutedError = 126
    CommandNotFoundError = 127
    InvalidExitArgumentError = 128
    FatalError = 129
    SigTerm = 130

    # These are the pypz specific error codes
    # =======================================

    StateServiceStartError = 110
    StateServiceShutdownError = 111
    StateResourceCreationError = 112
    StateResourcesDeletionError = 113
    StateOperationInitError = 114
    StateOperationError = 115
    StateOperationShutdownError = 116



# =============================================================================
# Copyright (c) 2024 by Laszlo Anka. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================
import os
import signal
import socket
import sys
import traceback

from pypz.executors.commons import ExecutionMode
from pypz.executors.commons import ExitCodes
from pypz.core.commons.utils import SynchronizedReference, TemplateResolver
from pypz.core.specs.operator import Operator
from pypz.core.specs.plugin import ExtendedPlugin, Plugin
from pypz.executors.operator.context import ExecutionContext
from pypz.executors.operator.signals import BaseSignal, SignalNoOp, SignalKill, SignalServicesStart, SignalShutdown, \
    SignalResourcesCreation, SignalResourcesDeletion, SignalError, SignalOperationInit, SignalServicesStop, \
    SignalOperationStart, SignalOperationStop
from pypz.executors.operator.states import State, StateEntry, StateKilled, StateOperationInit, StateOperationShutdown, \
    StateOperationRunning, StateResourceCreation, StateResourceDeletion, StateServiceStart, StateServiceShutdown
from pypz.version import PROJECT_VERSION


class OperatorExecutor:
    """
    This class has the purpose of executing an Operator along with its nested
    plugins. The execution is based on a state machine, where each state
    is responsible to execute specific entities' corresponding methods.

    :param operator: the actual operator instance to execute
    :param handle_interrupts: if True, then the execution can be interrupted by system signals
    """

    def __init__(self, operator: Operator, handle_interrupts: bool = True):

        # Initializing shutdown hook and signal handling as early as possible to prevent
        # the case that the signal has been sent during startup, but missed by the executors
        if handle_interrupts:
            signal.signal(signal.SIGTERM, self.interrupt)
            signal.signal(signal.SIGINT, self.interrupt)

        self.__operator: Operator = operator
        """
        The actual operator instance, which shall be processed by the state machine
        """

        self.__context: ExecutionContext | None = None
        """
        The stored execution context for this executor
        """

        self.__priority_signal: SynchronizedReference[BaseSignal] = SynchronizedReference(SignalNoOp())
        """
        Reference to the priority signal.
        Note that this atomic variable was necessary, because there might be functionality that attempts
        to change the current signal. Direct changing is not really a lucky choice if changing from a separate
        thread is allowed (e.g. from shutdown hook). Therefore all the external entities are allowed to change
        the priority signal
        """

        self.__is_running: SynchronizedReference[bool] = SynchronizedReference(False)
        """
        Flag that signalizes, whether the state machine is running
        """

        # ======= State declarations ========

        self.__state_entry: State | None = None
        self.__state_killed: State | None = None
        self.__state_operation_init: State | None = None
        self.__state_operation_running: State | None = None
        self.__state_operation_shutdown: State | None = None
        self.__state_resource_creation: State | None = None
        self.__state_resource_deletion: State | None = None
        self.__state_service_start: State | None = None
        self.__state_service_shutdown: State | None = None

        self.__current_state: State | None = None
        self.__current_signal: BaseSignal | None = None

        # Parameter handling
        # ==================

        # Resolve runtime templates as well for plugins
        for name, value in self.__operator.get_protected().get_parameters().items():
            self.__operator.set_parameter(name, TemplateResolver("$(", ")").resolve(value))

        for plugin in self.__operator.get_protected().get_nested_instances().values():
            for name, value in plugin.get_protected().get_parameters().items():
                plugin.set_parameter(name, TemplateResolver("$(", ")").resolve(value))

        # This is the point, where required parameters shall be checked, before continue
        missing_required_parameters = self.__operator.get_missing_required_parameters()

        if 0 < len(missing_required_parameters):
            raise LookupError(f"[{self.__operator.get_full_name()}] "
                              f"Missing required parameters: {missing_required_parameters}")

        # Mock instance checking
        # ======================
        # Mocked operators cannot be executed
        if "mocked" in operator.__class__.__dict__:
            raise PermissionError(f"[{operator.get_full_name()}] Mock operator cannot be executed")

        # If a non-mock operator has a mock plugin, then something went wrong
        for plugin in self.__operator.get_protected().get_nested_instances().values():
            if "mocked" in plugin.__class__.__dict__:
                raise PermissionError(f"[{plugin.get_full_name()}] Mock plugin in regular operator")

    # ========= public methods ==========

    def is_running(self):
        return self.__is_running.get()

    def get_current_state(self):
        return self.__current_state

    def get_current_signal(self):
        return self.__current_signal

    def execute(self, exec_mode: ExecutionMode = ExecutionMode.Standard) -> int:
        """
        Initializes and triggers execution. Note that, since every time, if this method is
        called, the persisted state of the execution and of its context is re-initialized,
        hence this method is idempotent.

        :param exec_mode: the run mode if the execution (refer to ExecutorRunMode)
        :return: exit code (refer to ExitCodes)
        """

        if ExecutionMode.Skip == exec_mode:
            return 0

        try:
            # ========= Initialize instances =========

            self.__initialize(exec_mode)

            # ========= Start the state machine =========

            try:
                self.__operator.get_logger().debug(f"Host: {os.getenv('PYPZ_NODE_NAME', socket.gethostname())}")

                self.__operator.get_logger().debug(f"Version: {PROJECT_VERSION}")

                self.__operator.get_logger().debug("Starting state machine ...")

                self.__operator.get_logger().debug("Run mode: %s", exec_mode.name)

                self.__is_running.set(True)

                while not isinstance(self.__current_signal, SignalKill):
                    self.__current_signal = self.__current_state.on_execute()

                    if not isinstance(self.__priority_signal.get(), SignalNoOp):
                        self.__operator.get_logger().debug("Priority signal caught: %s",
                                                           self.__priority_signal.get().__class__.__name__)
                        self.__current_signal = self.__priority_signal.get()
                        self.__priority_signal.set(SignalNoOp())

                    self.__current_state = self.__current_state.on_signal_handling(self.__current_signal)

            except:  # noqa: E722
                self.__context.set_exit_code(ExitCodes.FatalError)
                self.__operator.get_logger().error(traceback.format_exc())
            finally:
                self.__is_running.set(False)

                self.__operator.get_logger().debug("Shutting down state machine ...")
                self.__state_entry.shutdown()
                self.__state_killed.shutdown()
                self.__state_operation_init.shutdown()
                self.__state_operation_shutdown.shutdown()
                self.__state_operation_running.shutdown()
                self.__state_resource_creation.shutdown()
                self.__state_resource_deletion.shutdown()
                self.__state_service_start.shutdown()
                self.__state_service_shutdown.shutdown()

                try:
                    self.__context.for_each_plugin_objects_with_type(
                        ExtendedPlugin, lambda plugin: plugin.get_protected().post_execution())
                except:  # noqa: E722
                    self.__context.for_each_plugin_objects_with_type(
                        ExtendedPlugin, lambda plugin: plugin.get_protected().on_error())
                    raise
        except:  # noqa: E722
            # Catching exceptions not handled at this point
            traceback.print_exc(file=sys.stderr)
            self.__context.set_exit_code(ExitCodes.FatalError)

        if ExitCodes.NoError != self.__context.get_exit_code():
            print("[ERROR] Error occurred during the execution. Use logger plugin for details.", file=sys.stderr)

        return self.__context.get_exit_code().value

    # ========= private methods ==========

    def __initialize(self, exec_mode: ExecutionMode = ExecutionMode.Standard) -> None:
        """
        Initializes the execution states and context.

        :param exec_mode: the run mode if the execution (refer to ExecutorRunMode)
        """

        # ========= Initialize context =========

        self.__context: ExecutionContext = ExecutionContext(self.__operator, exec_mode)

        try:
            # Addons shall be initialized as early as possible to cover the most part
            # of the execution
            self.__context.for_each_plugin_objects_with_type(
                ExtendedPlugin, lambda plugin: plugin.get_protected().pre_execution())
        except:  # noqa: E722
            self.__context.for_each_plugin_objects_with_type(ExtendedPlugin,
                                                             lambda plugin: plugin.get_protected().on_error())
            raise

        # ========= Initialize state machine =========

        # State transitions init
        # ======================

        self.__state_entry = StateEntry(self.__context)
        self.__state_killed = StateKilled(self.__context)
        self.__state_operation_init = StateOperationInit(self.__context)
        self.__state_operation_shutdown = StateOperationShutdown(self.__context)
        self.__state_operation_running = StateOperationRunning(self.__context)
        self.__state_resource_creation = StateResourceCreation(self.__context)
        self.__state_resource_deletion = StateResourceDeletion(self.__context)
        self.__state_service_start = StateServiceStart(self.__context)
        self.__state_service_shutdown = StateServiceShutdown(self.__context)

        self.__state_entry.set_transition(SignalServicesStart, self.__state_service_start)
        self.__state_entry.set_transition(SignalShutdown, self.__state_killed)

        self.__state_service_start.set_transition(SignalResourcesCreation, self.__state_resource_creation)
        self.__state_service_start.set_transition(SignalResourcesDeletion, self.__state_resource_deletion)
        self.__state_service_start.set_transition(SignalError, self.__state_service_shutdown)
        self.__state_service_start.set_transition(SignalShutdown, self.__state_service_shutdown)

        self.__state_resource_creation.set_transition(SignalOperationInit, self.__state_operation_init)
        self.__state_resource_creation.set_transition(SignalError, self.__state_resource_deletion)
        self.__state_resource_creation.set_transition(SignalServicesStop, self.__state_service_shutdown)
        self.__state_resource_creation.set_transition(SignalShutdown, self.__state_resource_deletion)

        self.__state_operation_init.set_transition(SignalOperationStart, self.__state_operation_running)
        self.__state_operation_init.set_transition(SignalError, self.__state_operation_shutdown)
        self.__state_operation_init.set_transition(SignalShutdown, self.__state_operation_shutdown)

        self.__state_operation_running.set_transition(SignalOperationStop, self.__state_operation_shutdown)
        self.__state_operation_running.set_transition(SignalError, self.__state_operation_shutdown)
        self.__state_operation_running.set_transition(SignalShutdown, self.__state_operation_shutdown)

        self.__state_operation_shutdown.set_transition(SignalResourcesDeletion, self.__state_resource_deletion)
        self.__state_operation_shutdown.set_transition(SignalServicesStop, self.__state_service_shutdown)
        self.__state_operation_shutdown.set_transition(SignalError, self.__state_resource_deletion)

        self.__state_resource_deletion.set_transition(SignalServicesStop, self.__state_service_shutdown)
        self.__state_resource_deletion.set_transition(SignalError, self.__state_service_shutdown)

        self.__state_service_shutdown.set_transition(SignalError, self.__state_killed)
        self.__state_service_shutdown.set_transition(SignalKill, self.__state_killed)

        self.__current_state = self.__state_entry
        self.__current_signal = SignalNoOp()

    def interrupt(self, signal_number=None, current_stack=None):
        if self.__is_running.get():
            self.__operator.get_logger().debug("Processing interrupt signal ...")

            # Current state shutdown shall be called here to interrupt scheduling and cancel futures
            self.__current_state.shutdown()

            # Invoking plugins' on_interrupt() method
            try:
                self.__context.for_each_plugin_objects_with_type(
                    Plugin, lambda plugin: plugin.get_protected().on_interrupt(signal_number))
            except:  # noqa: E722
                # Ignore exception to be able to proceed with the shutdown
                traceback.print_exc(file=sys.stderr)
                pass

            # Invoking operator's on_interrupt() method
            try:
                self.__operator.get_protected().on_interrupt(signal_number)
            except:  # noqa: E722
                # Ignore exception to be able to proceed with the shutdown
                traceback.print_exc(file=sys.stderr)
                pass

            self.__context.set_exit_code(ExitCodes.SigTerm)
            self.__priority_signal.set(SignalShutdown())
